/******************************************************************************
 *
 * Source file for harnessd (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <harnessd.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Parses the command line options and configuration files, and runs the kernel
 * as daemon or program depending on the command line options and configuration
 * files.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = EXIT_SUCCESS on success or EXIT_FAILURE on any error.
 */
#undef  __FUNC__
#define __FUNC__ "main"
int main (int    argc,
          char **argv,
          char **envp);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Parses the command line options and configuration files, and runs the kernel
 * as daemon or program depending on the command line options and configuration
 * files.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = EXIT_SUCCESS on success or EXIT_FAILURE on any error.
 */
#undef  __FUNC__
#define __FUNC__ "main"
int main (int    argc,
          char **argv,
          char **envp) {
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Check envp parameter. */
  if (NULL == envp) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(envp parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
#endif /* DEBUG */
  /* Initialize options. */
  if (0 != harnessd_options_init(argc, (const char**)argv,
                                       (const char**)envp)) {
    HARNESS_LOG((HARNESS_WARN(unable to initialize options)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  /* Run the kernel. */
  if (0 != harness_main2()) {
    HARNESS_LOG((HARNESS_WARN(unable to run the kernel)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    /* Finalize options. */
    if (0 != harnessd_options_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize options)))
    }
    return EXIT_FAILURE;
  }
  /* Finalize options. */
  if (0 != harnessd_options_fini()) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize options)))
    fprintf(stdout, "ERROR: An internal error has occurred.\n");
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
