/******************************************************************************
 *
 * Source file for harnessd options (option parser) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Command line version output. */
#ifdef DEBUG
#define HARNESSD_OPTIONS_VERSION HARNESS_STRING(\
harnessd.HARNESSD_VERSION.debug (HARNESS kernel daemon in debug mode)\
 HARNESSD_PACKAGE_NAME-HARNESSD_PACKAGE_VERSION.HARNESSD_PACKAGE_RELEASE\n\
Copyright (c) 2002-2006 Oak Ridge National Laboratory.\n\
See COPYING in the source distribution top-level directory or package\n\
data directory (HARNESS_DATADIR) for license information.\n\
Report bugs to <HARNESSD_PACKAGE_BUGREPORT>.\n)
#else /* DEBUG */
#define HARNESSD_OPTIONS_VERSION HARNESS_STRING(\
harnessd.HARNESSD_VERSION (HARNESS kernel daemon)\
 HARNESSD_PACKAGE_NAME-HARNESSD_PACKAGE_VERSION.HARNESSD_PACKAGE_RELEASE\n\
Copyright (c) 2002-2006 Oak Ridge National Laboratory.\n\
See COPYING in the source distribution top-level directory or package\n\
data directory (HARNESS_DATADIR) for license information.\n\
Report bugs to <HARNESSD_PACKAGE_BUGREPORT>.\n)
#endif /* DEBUG */

/* Command line help output. */
#define HARNESSD_OPTIONS_HELP "\
Usage:\n\
  harnessd[.debug] [option[ |=][VALUE]] ..\n\
Environment:\n\
  HARNESSD_OPTIONS=\"[option[ |=][VALUE]] ..\"\n\
%s\
Configuration file format:\n\
  # Comment until end of line.\n\
  [option[ |=][VALUE]] .. \n\
  [option[ |=][VALUE]] .. \n\
  .. \n\
General options:\n\
  -h,  --help     Print out command line help. No option VALUE.\n\
  -v,  --version  Print out version information. No option VALUE.\n\
Plug-in/process options:\n\
  -e,  --exec     Execute a program with the name of VALUE in a\n\
                  child process at start-up. VALUE cannot be ommitted.\n\
  -l,  --load     Load a plug-in with the name of VALUE at start-up.\n\
                  Allow plug-in to load more than once and do not export\n\
                  its symbols. VALUE cannot be ommitted.\n\
  -le, --load-export\n\
                  Load a plug-in with the name of VALUE at start-up.\n\
                  Allow plug-in to load more than once and export its\n\
                  symbols. VALUE cannot be ommitted.\n\
  -lu, --load-unique\n\
                  Load a plug-in with the name of VALUE at start-up.\n\
                  Allow plug-in to load only once and do not export\n\
                  its symbols. VALUE cannot be ommitted.\n\
  -leu, --load-export-unique\n\
                  Load a plug-in with the name of VALUE at start-up.\n\
                  Allow plug-in to load only once and export its\n\
                  symbols. VALUE cannot be ommitted.\n\
  -f,  --file     Load a configuration file with the path of VALUE.\n\
                  VALUE cannot be ommitted.\n\
  -t,  --test     Loads the test plug-in. All other -l, --load and -e,\n\
                  --exec options are ignored. 0 <= VALUE <= 1,\n\
                  where 0 is default. Omitting VALUE defaults to 1.\n\
Kernel/forker options:\n\
  -b,  --buffer   Set the buffer size for forker-kernel buffers to VALUE.\n\
                  VALUE cannot be 0 or ommitted, where 4096 is default.\n\
  -D,  --daemon   Set the flag to run as daemon. 0 <= VALUE <= 1,\n\
                  where 1 is default. Omitting VALUE defaults to 1.\n\
  -d,  --debug    Set the flag to run in debug mode. 0 <= VALUE <= 1,\n\
                  where 1 is default in debug mode and 0 is default in.\n\
                  non-debug mode. Omitting VALUE defaults to 1.\n\
  -F,  --forker   Set the flag to run the forker. 0 <= VALUE <= 1,\n\
                  where 1 is default. Omitting VALUE defaults to 1.\n\
Thread pool options:\n\
  -m,  --minimum  Set the minimum thread count of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 5 is default.\n\
                  VALUE must be less than current maximum thread count.\n\
  -M,  --maximum  Set the maximum thread count of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 25 is default.\n\
                  VALUE must be greater than current minimum thread count.\n\
  -T,  --timeout  Set the exiting thread timeout of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 2 is default.\n\
  -q,  --queue    Set the maximum job count of the thread pool to VALUE.\n\
                  VALUE cannot be 0 or ommitted, where 1024 is default.\n\
  -w,  --wait     Set the waiting thread timeout of the thread pool to\n\
                  VALUE. VALUE cannot be ommitted, where 120 is default.\n\
"

#define HARNESSD_OPTIONS_HELP_FILES HARNESS_STRING(\
Directories:\n\
\40\40Plug-ins:\40 HARNESS_LOADDIR\n\
\40\40Programs:\40 HARNESS_EXECDIR\n\
\40\40Package:\40\40 HARNESS_DATADIR\n\
Configuration files:\n\
\40\40System:\40\40\40 HARNESS_SYSCONFDIR/HARNESS_CONFFILE\n\
\40\40User:\40\40\40\40\40 $HOME/HARNESS_USRCONFDIR/HARNESS_CONFFILE\n\
)

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harnessd.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Option parser data type. */
typedef struct {
  unsigned int     daemon;  /* daemon flag            */
  unsigned int     debug;   /* debug flag             */
  unsigned int     forker;  /* forker flag            */
  unsigned int     test;    /* test flag              */
  unsigned int     size;    /* forker buffer size     */
  struct {
    unsigned int   minimum; /* minimum thread count   */
    unsigned int   maximum; /* maximum thread count   */
    unsigned int   timeout; /* exiting thread timeout */
  }                pool;    /* thread pool options    */
  struct {
    unsigned int   maximum; /* maximum job count      */
    unsigned int   timeout; /* waiting thread timeout */
  }                jobs;    /* job queue options      */
  struct {
    unsigned int   count;   /* init options count     */
    char         **options; /* init options list      */
  }                init;    /* init options           */
} harnessd_options_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Parses a configuration file.
 *
 * path   = The configuration file path.
 * strict = The strict flag indicating if the file must exist.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
int harnessd_options_file_parse (const char    *path,
                                 unsigned int   strict,
                                 const char   **argv);

/*
 * Parses the environment options.
 *
 * envp   = The variable vector.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
int harnessd_options_envp_parse (const char **envp,
                                 const char **argv);

/*
 * Parses the command-line options.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
int harnessd_options_argv_parse (int          argc,
                                 const char **argv);

/*
 * Applies non-runtime options.
 *
 * argument = NULL
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int harnessd_options_apply ();

/*
 * Adds an option/value pair to the init options list.
 *
 * option = The option.
 * value  = The value.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harnessd_options_init_add (const char *option,
                               const char *value);

/*
 * Prints out a message to stdout and exits with an exit status.
 *
 * argv    = The argument vector (argument[0] is the program name).
 * format  = The message format.
 * message = The message.
 * status  = The exit status.
 */
void harnessd_options_exit (const char **argv,
                            const char  *format,
                            const char  *message,
                            int          status);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Option parser data. */
harnessd_options_data_t harnessd_options_data = {
  1,                              /* daemon flag            */
#ifdef DEBUG
  1,                              /* debug flag             */
#else /* DEBUG */
  0,                              /* debug flag             */
#endif /* DEBUG */
  1,                              /* forker flag            */
  0,                              /* test flag              */
  HARNESS_FORKER_BUFFER_SIZE,     /* forker buffer size     */
  {
    HARNESS_THREADS_POOL_MINIMUM, /* minimum thread count   */
    HARNESS_THREADS_POOL_MAXIMUM, /* maximum thread count   */
    HARNESS_THREADS_POOL_TIMEOUT  /* exiting thread timeout */
  },                              /* thread pool options    */
  {
    HARNESS_THREADS_JOBS_MAXIMUM, /* maximum job count      */
    HARNESS_THREADS_JOBS_TIMEOUT  /* waiting thread timeout */
  },                              /* job queue options      */
  {
    0,                            /* init options count     */
    NULL                          /* init options list      */
  }                               /* init options           */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the option parser.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_init"
int harnessd_options_init (const int    argc,
                           const char **argv,
                           const char **envp) {
  char         *home, *file, *path;
  unsigned int  homelen, filelen;
  HARNESS_LOG((HARNESS_INFO(option parser startup)))
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
  /* Check envp parameter. */
  if (NULL == envp) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(envp parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Parse system configuration file. */
  if (0 != harnessd_options_file_parse(
             HARNESS_STRING(HARNESS_SYSCONFDIR/HARNESS_CONFFILE), 0, argv)) {
    HARNESS_LOG((HARNESS_WARN(unable to parse system configuration file)))
    /* Finalize option parser. */
    if (0 != harnessd_options_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize option parser)))
    }
    return -1;
  }
  if (NULL != (home = getenv("HOME"))) {
    file    = HARNESS_STRING(/HARNESS_USRCONFDIR/HARNESS_CONFFILE);
    homelen = strlen(home);
    filelen = strlen(file) + 1;
    /* Allocate user configuration file path. */
    if (NULL == (path = malloc((homelen + filelen) * sizeof(char)))) {
      HARNESS_LOG((
        HARNESS_WARN(unable to allocate user configuration file path)))
      /* Finalize option parser. */
      if (0 != harnessd_options_fini()) {
        HARNESS_LOG((HARNESS_WARN(unable to finalize option parser)))
      }
      return -1;
    }
    /* Compose user configuration file path. */
    memcpy(path          , home, homelen);
    memcpy(path + homelen, file, filelen);
    /* Parse user configuration file. */
    if (0 != harnessd_options_file_parse(path, 0, argv)) {
      HARNESS_LOG((HARNESS_WARN(unable to parse user configuration file)))
      /* Deallocate user configuration file path. */
      free(path);
      /* Finalize option parser. */
      if (0 != harnessd_options_fini()) {
        HARNESS_LOG((HARNESS_WARN(unable to finalize option parser)))
      }
      return -1;
    }
    /* Deallocate user configuration file path. */
    free(path);
  }
  /* Parse environment options. */
  if (0 != harnessd_options_envp_parse(envp, argv)) {
    HARNESS_LOG((HARNESS_WARN(unable to parse environment options)))
    /* Finalize option parser. */
    if (0 != harnessd_options_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize option parser)))
    }
    return -1;
  }
  /* Parse command-line options. */
  if (0 != harnessd_options_argv_parse(argc, argv)) {
    HARNESS_LOG((HARNESS_WARN(unable to parse command-line options)))
    /* Finalize option parser. */
    if (0 != harnessd_options_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize option parser)))
    }
    return -1;
  }
  /* Apply non-runtime options. */
  if (0 != harnessd_options_apply(argv)) {
    HARNESS_LOG((HARNESS_WARN(unable to apply non-runtime options)))
    /* Finalize option parser. */
    if (0 != harnessd_options_fini()) {
      HARNESS_LOG((HARNESS_WARN(unable to finalize option parser)))
    }
    return -1;
  }
  return 0;
}

/*
 * Finalizes the option parser.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_fini"
int harnessd_options_fini () {
  unsigned int index;
  HARNESS_LOG((HARNESS_INFO(option parser shutdown)))
  /* Finalize init options list */
  for (index = 0; index < harnessd_options_data.init.count; index++) {
    if (NULL != harnessd_options_data.init.options[index]) {
      free(harnessd_options_data.init.options[index]);
      harnessd_options_data.init.options[index] = NULL;
    }
  }
  harnessd_options_data.init.count = 0;
  if (NULL != harnessd_options_data.init.options) {
    free(harnessd_options_data.init.options);
    harnessd_options_data.init.options = NULL;
  }
  return 0;
}

/*
 * Parses a configuration file.
 *
 * path   = The configuration file path.
 * strict = The strict flag indicating if the file must exist.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_file_parse"
int harnessd_options_file_parse (const char    *path,
                                 unsigned int   strict,
                                 const char   **argv) {
  int           file, bytes;
  char         *content, *options, *token, **argv1, **argv2;
  off_t         offset;
  unsigned int  length, index, mark, argc1, argc2;
#ifdef DEBUG
  /* Check path parameter. */
  if (NULL == path) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(path parameter is null)))
    return -1;
  }
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Open configuration file. */
  if (-1 == (file = open(path, O_RDONLY))) {
    if (1 == strict) {
      /* Exit with error message. */
      harnessd_options_exit(argv,
        "ERROR: Unable to open configuration file \'%s\'.\n",
        path, EXIT_FAILURE);
      return -1;
    } else {
      HARNESS_LOG((
        HARNESS_INFO(unable to open configuration file '%s'), path))
      return 0;
    }
  }
  /* Seek to end of configuration file. */
  if (-1 == (offset = lseek(file, 0, SEEK_END))) {
    HARNESS_LOG((HARNESS_WARN(unable to seek to end of configuration file)))
    /* Close configuration file. */
    if (0 != close(file)) {
      HARNESS_LOG((HARNESS_WARN(unable to close configuration file)))
    }
    return -1;
  }
  /* Calculate configuration file length. */
  length = offset;
  /* Seek to start of configuration file. */
  if (-1 == lseek(file, 0, SEEK_SET)) {
    HARNESS_LOG((
      HARNESS_WARN(unable to seek to start of configuration file)))
    /* Close configuration file. */
    if (0 != close(file)) {
      HARNESS_LOG((HARNESS_WARN(unable to close configuration file)))
    }
    return -1;
  }
  /* Check for empty configuration file. */
  if (0 == length) {
    /* Close configuration file. */
    if (0 != close(file)) {
      HARNESS_LOG((HARNESS_WARN(unable to close configuration file)))
    }
    return 0;
  }
  /* Allocate configuration file content. */
  if (NULL == (content = malloc(length + 1))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to allocate configuration file content)))
    /* Close configuration file. */
    if (0 != close(file)) {
      HARNESS_LOG((HARNESS_WARN(unable to close configuration file)))
    }
    return -1;
  }
  /* Read configuration file content. */
  for (index = 0; index < length; index += bytes) {
    /* Read some data. */
    if (-1 == (bytes = read(file, content + index, length - index))) {
      /* Check for other than interruption. */
      if ((errno != EINTR)&&(errno != EAGAIN)) {
        HARNESS_LOG((HARNESS_WARN(unable to read from configuration file)))
        /* Close configuration file. */
        if (0 != close(file)) {
          HARNESS_LOG((HARNESS_WARN(unable to close configuration file)))
        }
        return -1;
      }
    }
  }
  content[length] = '\0';
  /* Close configuration file. */
  if (0 != close(file)) {
    HARNESS_LOG((HARNESS_WARN(unable to close configuration file)))
    /* Deallocate configuration file content. */
    free(content);
    return -1;
  }
  /* Remove any comments from configuration file content. */
  for (index = 0; index < length; index++) {
    if ('#' == content[index]) {
      mark = index;
      for (index++; index < length && '\n' != content[index]; index++);
      memmove(content + mark,
              content + index, (length - index + 1) * sizeof(char));
      length -= (index - mark);
      index   = mark;
    }
  }
  /* Replace any new lines with spaces in configuration file content. */
  for (index = 0; index < length; index++) {
    if ('\n' == content[index]) {
      content[index] = ' ';
    }
  }
  argc1 = 1;
  /* Allocate argument vector. */
  if (NULL == (argv1 = (char**)malloc(sizeof(char*)))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate argument vector)))
    /* Deallocate configuration file content. */
    free(content);
    return -1;
  }
  /* Duplicate argument 0. */
  if (NULL == (argv1[0] = strdup(argv[0]))) {
    HARNESS_LOG((HARNESS_WARN(unable to duplicate argument 0)))
    /* Deallocate argument vector. */
    free(argv1);
    /* Deallocate configuration file content. */
    free(content);
    return -1;
  }
  /* Remove all leading spaces. */
  for (options = content; '\0' != options[0] && ' ' == options[0]; options++);
  while ('\0' != options[0]) {
    /* Check for ' quoting. */
    if ('\'' == options[0]) {
      /* Search for next ' quote that is not protected by \. */
      for (token = options + 1;
           ( '\0' != token[0])&&
           ((' '  != token[1])||('\'' != token[0])||('\\' == token[-1]))&&
           (('\0' != token[1])||('\'' != token[0])||('\\' == token[-1]));
           token++);
      /* Check for missing ' quote. */
      if ('\0' == token[0]) {
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        free(argv1);
        /* Deallocate configuration file content. */
        free(content);
        /* Exit with error message. */
        harnessd_options_exit(argv,
          "ERROR: Missing closing quote in configuration file '%s'.\n",
          path, EXIT_FAILURE);
      }
    /* Check for " quoting. */
    } else if ('\"' == options[0]) {
      /* Search for next " quote that is not protected by \. */
      for (token = options + 1;
           ( '\0' != token[0])&&
           ((' '  != token[1])||('\"' != token[0])||('\\' == token[-1]))&&
           (('\0' != token[1])||('\"' != token[0])||('\\' == token[-1]));
           token++);
      /* Check for missing " quote. */
      if ('\0' == token[0]) {
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        free(argv1);
        /* Deallocate configuration file content. */
        free(content);
        /* Exit with error message. */
        harnessd_options_exit(argv,
          "ERROR: Missing closing quote in configuration file '%s'.\n",
          path, EXIT_FAILURE);
      }
    } else {
      /* Search for next space. */
      for (token = options; '\0' != token[0] && ' ' != token[0]; token++);
    }
    /* Increase argument vector. */
    argc2 = argc1;
    argc1++;
    /* Reallocate argument vector. */
    if (NULL == (argv2 = (char**)realloc(argv1, argc1 * sizeof(char*)))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate argument vector)))
      /* Deallocate arguments. */
      while (0 != argc2) {
        argc2--;
          free(argv1[argc2]);
      }
      /* Deallocate argument vector. */
      free(argv1);
      /* Deallocate configuration file content. */
      free(content);
      return -1;
    }
    argv1 = argv2;
    /* Duplicate argument. */
    if (NULL == (argv1[argc2] = strndup(options, token - options))) {
      HARNESS_LOG((HARNESS_WARN(unable to duplicate argument)))
      /* Deallocate arguments. */
      while (0 != argc2) {
        argc2--;
        free(argv1[argc2]);
      }
      /* Deallocate argument vector. */
      free(argv1);
      /* Deallocate configuration file content. */
      free(content);
      return -1;
    }
    options = token;
    if ('\0' != options[0]) {
      options++;
    }
    /* Remove all leading spaces. */
    for (; '\0' != options[0] && ' ' == options[0]; options++);
  }
  /* Parse environment options. */
  if (0 != harnessd_options_argv_parse(argc1, (const char**)argv1)) {
    HARNESS_LOG((HARNESS_WARN(unable to parse configuration file options)))
    /* Deallocate arguments. */
    while (0 != argc1) {
      argc1--;
      free(argv1[argc1]);
    }
    /* Deallocate argument vector. */
    free(argv1);
    /* Deallocate configuration file content. */
    free(content);
    return -1;
  }
  /* Deallocate arguments. */
  while (0 != argc1) {
    argc1--;
    free(argv1[argc1]);
  }
  /* Deallocate argument vector. */
  free(argv1);
  /* Deallocate configuration file content. */
  free(content);
  return 0;
}

/*
 * Parses the environment options.
 *
 * envp   = The variable vector.
 * argv   = The argument vector (argument[0] is the program name).
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_envp_parse"
int harnessd_options_envp_parse (const char **envp,
                                 const char **argv) {
  char         **argv1,   **argv2;
  const char    *options,  *token;
  unsigned int   index, argc1, argc2;
#ifdef DEBUG
  /* Check envp parameter. */
  if (NULL == envp) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(envp parameter is null)))
    return -1;
  }
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 0; NULL != envp[index]; index++) {
    if (0 == strncmp(envp[index], "HARNESSD_OPTIONS=", 17)) {
      argc1   = 1;
      options = envp[index] + 17;
      /* Allocate argument vector. */
      if (NULL == (argv1 = (char**)malloc(sizeof(char*)))) {
        HARNESS_LOG((HARNESS_WARN(unable to allocate argument vector)))
        return -1;
      }
      /* Duplicate argument 0. */
      if (NULL == (argv1[0] = strdup(argv[0]))) {
        HARNESS_LOG((HARNESS_WARN(unable to duplicate argument 0)))
        /* Deallocate argument vector. */
        free(argv1);
        return -1;
      }
      /* Remove all leading spaces. */
      for (; '\0' != options[0] && ' ' == options[0]; options++);
      while ('\0' != options[0]) {
        /* Check for ' quoting. */
        if ('\'' == options[0]) {
          /* Search for next ' quote that is not protected by \. */
          for (token = options + 1;
               ( '\0' != token[0])&&
               ((' '  != token[1])||('\'' != token[0])||('\\' == token[-1]))&&
               (('\0' != token[1])||('\'' != token[0])||('\\' == token[-1]));
               token++);
          /* Check for missing ' quote. */
          if ('\0' == token[0]) {
            /* Deallocate arguments. */
            while (0 != argc1) {
              argc1--;
              free(argv1[argc1]);
            }
            /* Deallocate argument vector. */
            free(argv1);
            /* Exit with error message. */
            harnessd_options_exit(argv,
              "ERROR: Missing closing quote in %s variable.\n",
              "HARNESSD_OPTIONS", EXIT_FAILURE);
          }
        /* Check for " quoting. */
        } else if ('\"' == options[0]) {
          /* Search for next " quote that is not protected by \. */
          for (token = options + 1;
               ( '\0' != token[0])&&
               ((' '  != token[1])||('\"' != token[0])||('\\' == token[-1]))&&
               (('\0' != token[1])||('\"' != token[0])||('\\' == token[-1]));
               token++);
          /* Check for missing " quote. */
          if ('\0' == token[0]) {
            /* Deallocate arguments. */
            while (0 != argc1) {
              argc1--;
              free(argv1[argc1]);
            }
            /* Deallocate argument vector. */
            if (NULL != argv1) {
              free(argv1);
            }
            /* Exit with error message. */
            harnessd_options_exit(argv,
              "ERROR: Missing closing quote in %s variable.\n",
              "HARNESSD_OPTIONS", EXIT_FAILURE);
          }
        } else {
          /* Search for next space. */
          for (token = options; '\0' != token[0] && ' ' != token[0]; token++);
        }
        /* Increase argument vector. */
        argc2 = argc1;
        argc1++;
        /* Reallocate argument vector. */
        if (NULL == (argv2 = (char**)realloc(argv1, argc1 * sizeof(char*)))) {
          HARNESS_LOG((HARNESS_WARN(unable to reallocate argument vector)))
          /* Deallocate arguments. */
          while (0 != argc2) {
            argc2--;
            free(argv1[argc2]);
          }
          /* Deallocate argument vector. */
          free(argv1);
          return -1;
        }
        argv1 = argv2;
        /* Duplicate argument. */
        if (NULL == (argv1[argc2] = strndup(options, token - options))) {
          HARNESS_LOG((HARNESS_WARN(unable to duplicate argument)))
          /* Deallocate arguments. */
          while (0 != argc2) {
            argc2--;
            free(argv1[argc2]);
          }
          /* Deallocate argument vector. */
          free(argv1);
          return -1;
        }
        options = token;
        if ('\0' != options[0]) {
          options++;
        }
        /* Remove all leading spaces. */
        for (; '\0' != options[0] && ' ' == options[0]; options++);
      }
      /* Parse environment options. */
      if (0 != harnessd_options_argv_parse(argc1, (const char**)argv1)) {
        HARNESS_LOG((HARNESS_WARN(unable to parse environment options)))
        /* Deallocate arguments. */
        while (0 != argc1) {
          argc1--;
          free(argv1[argc1]);
        }
        /* Deallocate argument vector. */
        free(argv1);
        return -1;
      }
      /* Deallocate arguments. */
      while (0 != argc1) {
        argc1--;
        if (NULL != argv1[argc1]) {
          free(argv1[argc1]);
        }
      }
      /* Deallocate argument vector. */
      free(argv1);
    }
  }
  return 0;
}

/*
 * Parses the command-line options.
 *
 * argc   = The argument vector count.
 * argv   = The argument vector (argument[0] is the program name).
 * envp   = The variable vector.
 * return = 0 on success or -1 on any error with errno set appropriately.
 *          May exit with EXIT_SUCCESS if an option requires it and may exit
 *          with EXIT_FAILURE for unrecognized options and missing values.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_argv_parse"
int harnessd_options_argv_parse (int          argc,
                                 const char **argv) {
  unsigned int index;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  for (index = 1; index < argc; index++) {
    if        ((0 == strcmp (argv[index],  "-b"     ))||
               (0 == strcmp (argv[index], "--buffer"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        harnessd_options_data.size = atoi(argv[index]);
        /* Check range. */
        if (0 == harnessd_options_data.size) {
          /* Exit with error message. */
          harnessd_options_exit(argv,
                                "ERROR: Value for '%s' option cannot be 0.\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-b=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.size = atoi(argv[index] + 3);
      /* Check range. */
      if (0 == harnessd_options_data.size) {
        /* Exit with error message. */
        harnessd_options_exit(argv,
                              "ERROR: Value for '%s' option cannot be 0.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "--buffer=", 9)) {
      /* Check for value in argument. */
      if (9 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.size = atoi(argv[index] + 9);
      /* Check range. */
      if (0 == harnessd_options_data.size) {
        /* Exit with error message. */
        harnessd_options_exit(argv,
                              "ERROR: Value for '%s' option cannot be 0.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if ((0 == strcmp (argv[index],  "-D"     ))||
               (0 == strcmp (argv[index], "--daemon"))) {
      /* Check for next argument. */
      if ((index + 1 < argc)&&('-' != argv[index + 1][0])) {
        index++;
        if (0 == strcmp(argv[index], "0")) {
          harnessd_options_data.daemon = 0;
        } else if (0 == strcmp(argv[index], "1")) {
          harnessd_options_data.daemon = 1;
        } else {
          /* Exit with error message. */
          harnessd_options_exit(argv,
                                "ERROR: Illegal value for '%s' option .\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        harnessd_options_data.daemon = 1;
      }
    } else if ((0 == strcmp (argv[index],  "-D=0"     ))||
               (0 == strcmp (argv[index], "--daemon=0"))) {
      harnessd_options_data.daemon = 0;
    } else if ((0 == strcmp (argv[index],  "-D=1"     ))||
               (0 == strcmp (argv[index], "--daemon=1"))) {
      harnessd_options_data.daemon = 1;
    } else if ((0 == strcmp (argv[index],  "-D="     ))||
               (0 == strcmp (argv[index], "--daemon="))) {
      harnessd_options_data.daemon = 1;
    } else if ((0 == strncmp(argv[index],  "-D="     , 3))||
               (0 == strncmp(argv[index], "--daemon=", 9))) {
      /* Exit with error message. */
      harnessd_options_exit(argv, "ERROR: Illegal value for '%s' option .\n",
                            argv[index], EXIT_FAILURE);
    } else if ((0 == strcmp (argv[index],  "-d"    ))||
               (0 == strcmp (argv[index], "--debug"))) {
      /* Check for next argument. */
      if ((index + 1 < argc)&&('-' != argv[index + 1][0])) {
        index++;
        if (0 == strcmp(argv[index], "0")) {
          harnessd_options_data.debug = 0;
        } else if (0 == strcmp(argv[index], "1")) {
          harnessd_options_data.debug = 1;
        } else {
          /* Exit with error message. */
          harnessd_options_exit(argv,
                                "ERROR: Illegal value for '%s' option .\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        harnessd_options_data.debug = 1;
      }
    } else if ((0 == strcmp (argv[index],  "-d=0"    ))||
               (0 == strcmp (argv[index], "--debug=0"))) {
      harnessd_options_data.debug = 0;
    } else if ((0 == strcmp (argv[index],  "-d=1"    ))||
               (0 == strcmp (argv[index], "--debug=1"))) {
      harnessd_options_data.debug = 1;
    } else if ((0 == strcmp (argv[index],  "-d="    ))||
               (0 == strcmp (argv[index], "--debug="))) {
      harnessd_options_data.debug = 1;
    } else if ((0 == strncmp(argv[index],  "-d="    , 3))||
               (0 == strncmp(argv[index], "--debug=", 8))) {
      /* Exit with error message. */
      harnessd_options_exit(argv, "ERROR: Illegal value for '%s' option .\n",
                            argv[index], EXIT_FAILURE);
    } else if ((0 == strcmp (argv[index],  "-e"   ))||
               (0 == strcmp (argv[index], "--exec"))) {
      index++;
      if (index < argc) {
        /* Add to init options list. */
        if (0 != harnessd_options_init_add("exec", argv[index])) {
          HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
          return -1;
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index],  "-e="   , 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("exec", argv[index] + 3)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--exec=", 7)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("exec", argv[index] + 7)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ((0 == strcmp (argv[index],  "-f"   ))||
               (0 == strcmp (argv[index], "--file"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        /* Parse configuration file. */
        if (0 != harnessd_options_file_parse(argv[index], 1, argv)) {
          HARNESS_LOG((HARNESS_WARN(unable to parse configuration file)))
          return -1;
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index],  "-f="   , 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Parse configuration file. */
      if (0 != harnessd_options_file_parse(argv[index] + 3, 1, argv)) {
        HARNESS_LOG((HARNESS_WARN(unable to parse configuration file)))
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--file=", 7)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Parse configuration file. */
      if (0 != harnessd_options_file_parse(argv[index] + 7, 1, argv)) {
        HARNESS_LOG((HARNESS_WARN(unable to parse configuration file)))
        return -1;
      }
    } else if ((0 == strcmp (argv[index],  "-F"     ))||
               (0 == strcmp (argv[index], "--forker"))) {
      /* Check for next argument. */
      if ((index + 1 < argc)&&('-' != argv[index + 1][0])) {
        index++;
        if (0 == strcmp(argv[index], "0")) {
          harnessd_options_data.forker = 0;
        } else if (0 == strcmp(argv[index], "1")) {
          harnessd_options_data.forker = 1;
        } else {
          /* Exit with error message. */
          harnessd_options_exit(argv,
                                "ERROR: Illegal value for '%s' option .\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        harnessd_options_data.forker = 1;
      }
    } else if ((0 == strcmp (argv[index],  "-F=0"     ))||
               (0 == strcmp (argv[index], "--forker=0"))) {
      harnessd_options_data.forker = 0;
    } else if ((0 == strcmp (argv[index],  "-F=1"     ))||
               (0 == strcmp (argv[index], "--forker=1"))) {
      harnessd_options_data.forker = 1;
    } else if ((0 == strcmp (argv[index],  "-F="     ))||
               (0 == strcmp (argv[index], "--forker="))) {
      harnessd_options_data.forker = 1;
    } else if ((0 == strncmp(argv[index],  "-F="     , 3))||
               (0 == strncmp(argv[index], "--forker=", 9))) {
      /* Exit with error message. */
      harnessd_options_exit(argv, "ERROR: Illegal value for '%s' option .\n",
                            argv[index], EXIT_FAILURE);
    } else if ((0 == strcmp (argv[index],  "-h"   ))||
               (0 == strcmp (argv[index], "--help"))) {
      /* Exit with help message. */
      harnessd_options_exit(argv, HARNESSD_OPTIONS_HELP,
                            HARNESSD_OPTIONS_HELP_FILES, EXIT_SUCCESS);

    } else if ((0 == strcmp (argv[index],  "-l"   ))||
               (0 == strcmp (argv[index], "--load"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        /* Add to init options list. */
        if (0 != harnessd_options_init_add("load", argv[index])) {
          HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
          return -1;
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index],  "-l="   , 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load", argv[index] + 3)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--load=", 7)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load", argv[index] + 7)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ((0 == strcmp (argv[index],  "-le"           ))||
               (0 == strcmp (argv[index], "--load-export"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        /* Add to init options list. */
        if (0 != harnessd_options_init_add("load-export", argv[index])) {
          HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
          return -1;
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index],  "-le="   , 4)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load-export", argv[index] + 4)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--load-export=", 14)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load-export", argv[index] + 14)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ((0 == strcmp (argv[index],  "-lu"         ))||
               (0 == strcmp (argv[index], "--load-unique"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        /* Add to init options list. */
        if (0 != harnessd_options_init_add("load-unique", argv[index])) {
          HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
          return -1;
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index],  "-lu="   , 4)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load-unique", argv[index] + 4)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--load-unique=", 14)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load-unique", argv[index] + 14)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ((0 == strcmp (argv[index],  "-leu"               ))||
               (0 == strcmp (argv[index], "--load-export-unique"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        /* Add to init options list. */
        if (0 != harnessd_options_init_add("load-export-unique", argv[index])) {
          HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
          return -1;
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index],  "-leu="   , 5)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load-export-unique",
                                         argv[index] + 5)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if ( 0 == strncmp(argv[index], "--load-export-unique=", 21)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      /* Add to init options list. */
      if (0 != harnessd_options_init_add("load-export-unique",
                                         argv[index] + 21)) {
        HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
        return -1;
      }
    } else if  ((0 == strcmp (argv[index],  "-m"      ))||
                (0 == strcmp (argv[index], "--minimum"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        harnessd_options_data.pool.minimum = atoi(argv[index]);
        /* Check range. */
        if (harnessd_options_data.pool.minimum >
            harnessd_options_data.pool.maximum) {
          /* Exit with error message. */
          harnessd_options_exit(argv, "ERROR: Value for '%s' is too big.\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-m=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.pool.minimum = atoi(argv[index] + 3);
      /* Check range. */
      if (harnessd_options_data.pool.minimum >
          harnessd_options_data.pool.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Value for '%s' is too big.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "--minimum=", 10)) {
      /* Check for value in argument. */
      if (10 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.pool.minimum = atoi(argv[index] + 10);
      /* Check range. */
      if (harnessd_options_data.pool.minimum >
          harnessd_options_data.pool.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv,
                              "ERROR: Value for '%s' is too big.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if  ((0 == strcmp (argv[index],  "-M"      ))||
                (0 == strcmp (argv[index], "--maximum"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        harnessd_options_data.pool.maximum = atoi(argv[index]);
        /* Check range. */
        if (0 == harnessd_options_data.pool.maximum) {
          /* Exit with error message. */
          harnessd_options_exit(argv,
                                "ERROR: Value for '%s' option cannot be 0.\n",
                                argv[index - 1], EXIT_FAILURE);
        }
        if (harnessd_options_data.pool.minimum >
            harnessd_options_data.pool.maximum) {
          /* Exit with error message. */
          harnessd_options_exit(argv, "ERROR: Value for '%s' is too small.\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-M=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.pool.maximum = atoi(argv[index] + 3);
      /* Check range. */
      if (0 == harnessd_options_data.pool.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv,
                              "ERROR: Value for '%s' option cannot be 0.\n",
                              argv[index], EXIT_FAILURE);
      }
      if (harnessd_options_data.pool.minimum >
          harnessd_options_data.pool.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Value for '%s' is too small.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "--maximum=", 10)) {
      /* Check for value in argument. */
      if (10 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.pool.maximum = atoi(argv[index] + 10);
      /* Check range. */
      if (0 == harnessd_options_data.pool.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv,
                              "ERROR: Value for '%s' option cannot be 0.\n",
                              argv[index], EXIT_FAILURE);
      }
      if (harnessd_options_data.pool.minimum >
          harnessd_options_data.pool.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Value for '%s' is too small.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if  ((0 == strcmp (argv[index],  "-q"    ))||
                (0 == strcmp (argv[index], "--queue"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        harnessd_options_data.jobs.maximum = atoi(argv[index]);
        /* Check range. */
        if (0 == harnessd_options_data.jobs.maximum) {
          /* Exit with error message. */
          harnessd_options_exit(argv,
                                "ERROR: Value for '%s' option cannot be 0.\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-q=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.jobs.maximum = atoi(argv[index] + 3);
      /* Check range. */
      if (0 == harnessd_options_data.jobs.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv,
                              "ERROR: Value for '%s' option cannot be 0.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "--queue=", 8)) {
      /* Check for value in argument. */
      if (8 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.jobs.maximum = atoi(argv[index] + 8);
      /* Check range. */
      if (0 == harnessd_options_data.jobs.maximum) {
        /* Exit with error message. */
        harnessd_options_exit(argv,
                              "ERROR: Value for '%s' option cannot be 0.\n",
                              argv[index], EXIT_FAILURE);
      }
    } else if ((0 == strcmp (argv[index],  "-t"   ))||
               (0 == strcmp (argv[index], "--test"))) {
      /* Check for next argument. */
      if ((index + 1 < argc)&&('-' != argv[index + 1][0])) {
        index++;
        if (0 == strcmp(argv[index], "0")) {
          harnessd_options_data.test = 0;
        } else if (0 == strcmp(argv[index], "1")) {
          harnessd_options_data.test = 1;
        } else {
          /* Exit with error message. */
          harnessd_options_exit(argv,
                                "ERROR: Illegal value for '%s' option .\n",
                                argv[index - 1], EXIT_FAILURE);
        }
      } else {
        harnessd_options_data.test = 1;
      }
    } else if ((0 == strcmp (argv[index],  "-t=0"   ))||
               (0 == strcmp (argv[index], "--test=0"))) {
      harnessd_options_data.test = 0;
    } else if ((0 == strcmp (argv[index],  "-t=1"   ))||
               (0 == strcmp (argv[index], "--test=1"))) {
      harnessd_options_data.test = 1;
    } else if ((0 == strcmp (argv[index],  "-t="   ))||
               (0 == strcmp (argv[index], "--test="))) {
      harnessd_options_data.test = 1;
    } else if ((0 == strncmp(argv[index],  "-t="   , 3))||
               (0 == strncmp(argv[index], "--test=", 7))) {
      /* Exit with error message. */
      harnessd_options_exit(argv, "ERROR: Illegal value for '%s' option .\n",
                            argv[index], EXIT_FAILURE);
    } else if  ((0 == strcmp (argv[index],  "-T"      ))||
                (0 == strcmp (argv[index], "--timeout"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        harnessd_options_data.pool.timeout = atoi(argv[index]);
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-T=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.pool.timeout = atoi(argv[index] + 3);
    } else if ( 0 == strncmp(argv[index], "--timeout=", 10)) {
      /* Check for value in argument. */
      if (10 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.pool.timeout = atoi(argv[index] + 10);
    } else if ((0 == strcmp (argv[index],  "-v"      ))||
               (0 == strcmp (argv[index], "--version"))) {
      /* Exit with version message. */
      harnessd_options_exit(argv, "%s", HARNESSD_OPTIONS_VERSION, EXIT_SUCCESS);

    } else if  ((0 == strcmp (argv[index],  "-w"   ))||
                (0 == strcmp (argv[index], "--wait"))) {
      /* Check for next argument. */
      index++;
      if (index < argc) {
        harnessd_options_data.jobs.timeout = atoi(argv[index]);
      } else {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value for '%s' option.\n",
                              argv[index - 1], EXIT_FAILURE);
      }
    } else if ( 0 == strncmp(argv[index], "-w=", 3)) {
      /* Check for value in argument. */
      if (3 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.jobs.timeout = atoi(argv[index] + 3);
    } else if ( 0 == strncmp(argv[index], "--wait=", 7)) {
      /* Check for value in argument. */
      if (7 == strlen(argv[index])) {
        /* Exit with error message. */
        harnessd_options_exit(argv, "ERROR: Missing value in '%s' option.\n",
                              argv[index], EXIT_FAILURE);
      }
      harnessd_options_data.jobs.timeout = atoi(argv[index] + 7);
    } else {
      /* Exit with error message. */
      harnessd_options_exit(argv, "ERROR: Unknown option '%s'.\n",
                            argv[index], EXIT_FAILURE);
    }
  }
  return 0;
}

/*
 * Applies non-runtime options.
 *
 * argument = NULL
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_apply"
int harnessd_options_apply (const char **argv) {
  char         *argv0;
  unsigned int  length, index;
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    return -1;
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    return -1;
  }
#endif /* DEBUG */
#ifdef DEBUG
  /* Check the debug flag. */
  if (0 == harnessd_options_data.debug) {
    length = strlen(argv[0]) - 6;
    /* Allocate argv[0] duplicate. */
    if (NULL == (argv0 = malloc((length + 1)* sizeof(char)))) {
      HARNESS_LOG((HARNESS_WARN(unable to allocate argv[0] duplicate)))
      return -1;
    }
    memcpy(argv0, argv[0], length);
    argv0[length] = 0;
    /* Execute non-debug version. */
    execv(argv0, (char**)argv);
    HARNESS_LOG((HARNESS_WARN(unable to execute non-debug version)))
    return -1;
  }
#else /* DEBUG */
  /* Check the debug flag. */
  if (1 == harnessd_options_data.debug) {
    length = strlen(argv[0]);
    /* Allocate argv[0] duplicate. */
    if (NULL == (argv0 = malloc((length  + 7)* sizeof(char)))) {
      return -1;
    }
    memcpy(argv0         , argv[0] , length);
    memcpy(argv0 + length, ".debug", 7);
    execv(argv0, (char**)argv);
    return -1;
  }
#endif /* DEBUG */
  /* Check the test flag. */
  if (1 == harnessd_options_data.test) {
    /* Finalize init options list */
    for (index = 0; index < harnessd_options_data.init.count; index++) {
      if (NULL != harnessd_options_data.init.options[index]) {
        free(harnessd_options_data.init.options[index]);
        harnessd_options_data.init.options[index] = NULL;
      }
    }
    harnessd_options_data.init.count = 0;
    if (NULL != harnessd_options_data.init.options) {
      free(harnessd_options_data.init.options);
      harnessd_options_data.init.options = NULL;
    }
    /* Add to init options list. */
    if (0 != harnessd_options_init_add("load",
         HARNESS_STRING(libharness-test.LIBHARNESS_TEST_VERSION_CURRENT))) {
      HARNESS_LOG((HARNESS_WARN(unable to add to init options list)))
      return -1;
    }
  }
  /* Set the daemon flag. */
  if (0 != harness_daemon_set(harnessd_options_data.daemon)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the daemon flag)))
    return -1;
  }
  /* Set the forker flag. */
  if (0 != harness_forker_set(harnessd_options_data.forker)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the forker flag)))
    return -1;
  }
  /* Set the forker buffer size. */
  if (0 != harness_forker_size_set(harnessd_options_data.size)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the forker buffer size)))
    return -1;
  }
  /* Set the minimum thread count. */
  if (0 !=
      harness_threads_pool_minimum_set(harnessd_options_data.pool.minimum)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the minimum thread count)))
    return -1;
  }
  /* Set the maximum thread count. */
  if (0 !=
      harness_threads_pool_maximum_set(harnessd_options_data.pool.maximum)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the maximum thread count)))
    return -1;
  }
  /* Set the exiting thread timeout. */
  if (0 !=
      harness_threads_pool_timeout_set(harnessd_options_data.pool.timeout)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the exiting thread timeout)))
    return -1;
  }
  /* Set the maximum job count. */
  if (0 !=
      harness_threads_jobs_maximum_set(harnessd_options_data.jobs.maximum)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the maximum job count)))
    return -1;
  }
  /* Set the waiting thread timeout. */
  if (0 !=
      harness_threads_jobs_timeout_set(harnessd_options_data.jobs.timeout)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the waiting thread timeout)))
    return -1;
  }
  /* Set the kernel init options. */
  if (0 != harness_kernel_options_set(harnessd_options_data.init.options)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the kernel init options)))
    return -1;
  }
  /* Set the kernel block flag. */
  if (0 != harness_kernel_block_set(1)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the kernel block flag.)))
    return -1;
  }
  return 0;
}

/*
 * Adds an option/value pair to the init options list.
 *
 * option = The option.
 * value  = The value.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_init_add"
int harnessd_options_init_add (const char *option,
                               const char *value) {
  void         *options;
  unsigned int  index;
#ifdef DEBUG
  /* Check option parameter. */
  if (NULL == option) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(option parameter is null)))
    return -1;
  }
  /* Check value parameter. */
  if (NULL == value) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(value parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  /* Increase init options list. */
  index = harnessd_options_data.init.count;
  harnessd_options_data.init.count += 2;
  /* Reallocate init options list. */
  if (NULL == (options = realloc(harnessd_options_data.init.options,
                                (harnessd_options_data.init.count + 1) *
                          sizeof(harnessd_options_data.init.options[0])))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate init options list)))
    return -1;
  }
  harnessd_options_data.init.options = options;
  /* Duplicate option. */
  if (NULL == (harnessd_options_data.init.options[index] = strdup(option))) {
    HARNESS_LOG((HARNESS_WARN(unable to duplicate option)))
    /* Decrease init options list. */
    harnessd_options_data.init.count -= 2;
    /* Reallocate init options list. */
    if (0 == harnessd_options_data.init.count) {
      free(harnessd_options_data.init.options);
      harnessd_options_data.init.options = NULL;
    } else if (NULL == (options = 
                        realloc(harnessd_options_data.init.options,
                               (harnessd_options_data.init.count + 1) *
                         sizeof(harnessd_options_data.init.options[0])))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate init options list)))
    } else {
      harnessd_options_data.init.options = options;
    }
    return -1;
  }
  /* Duplicate value. */
  if (NULL == (harnessd_options_data.init.options[index + 1] = strdup(value))) {
    HARNESS_LOG((HARNESS_WARN(unable to duplicate value)))
    /* Decrease init options list. */
    harnessd_options_data.init.count -= 2;
    /* Reallocate init options list. */
    if (0 == harnessd_options_data.init.count) {
      free(harnessd_options_data.init.options);
      harnessd_options_data.init.options = NULL;
    } else if (NULL == (options =
                        realloc(harnessd_options_data.init.options,
                               (harnessd_options_data.init.count + 1) *
                         sizeof(harnessd_options_data.init.options[0])))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate init options list)))
    } else {
      harnessd_options_data.init.options = options;
    }
    return -1;
  }
  /* Terminate init options list. */
  harnessd_options_data.init.options[index + 2] = NULL;
  return 0;
}

/*
 * Prints out a message to stdout and exits with an exit status.
 *
 * argv    = The argument vector (argument[0] is the program name).
 * format  = The message format.
 * message = The message.
 * status  = The exit status.
 */
#undef  __FUNC__
#define __FUNC__ "harnessd_options_exit"
void harnessd_options_exit (const char **argv,
                            const char  *format,
                            const char  *message,
                            int          status) {
#ifdef DEBUG
  /* Check argv parameter. */
  if (NULL == argv) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
  /* Check argv[0] parameter. */
  if (NULL == argv[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(argv[0] parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
  /* Check format parameter. */
  if (NULL == format) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(format parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
  /* Check message parameter. */
  if (NULL == message) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(message parameter is null)))
    fprintf(stdout, "ERROR: An internal error has occurred.");
    exit(EXIT_FAILURE);
  }
#endif /* DEBUG */
  /* Finalize option parser. */
  if (0 != harnessd_options_fini()) {
    HARNESS_LOG((HARNESS_WARN(unable to finalize option parser)))
    exit(EXIT_FAILURE);
  }
  fprintf(stdout, format, message);
  if (EXIT_FAILURE == status) {
    fprintf(stdout, "Try `%s --help` for more information.\n", argv[0]);
  }
  exit(status);
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
