/******************************************************************************
 *
 * Source file for rmix interface (object interface) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Creates an object interface.
 *
 * interface = The object interface to create (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_create"
int rmix_interface_create (rmix_interface_t **interface) {
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate object interface. */
  if (NULL == (*interface = malloc(sizeof(rmix_interface_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate object interface)))
    errno = errno2;
    return -1;
  }
  (*interface)->count      = 0;
  (*interface)->methods = NULL;
  return 0;
}

/*
 * Creates an object interface.
 *
 * interface = The object interface to create (return).
 * count     = The object method descriptor list size.
 * methods   = The object method descriptor list.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_create2"
int rmix_interface_create2 (rmix_interface_t    **interface,
                            unsigned int          count,
                            const rmix_method_t  *methods) {
  unsigned int index;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check methods parameter. */
  if ((NULL == methods)&&(0 != count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check count parameter. */
  if ((0 == count)&&(NULL != methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate object interface. */
  if (NULL == (*interface = malloc(sizeof(rmix_interface_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate object interface)))
    errno = errno2;
    return -1;
  }
  (*interface)->count = 0;
  /* Check count parameter. */
  if (0 == count) {
    (*interface)->methods = NULL;
    return 0;
  }
  /* Allocate object method descriptor list. */
  if (NULL == ((*interface)->methods =
               malloc(sizeof(rmix_method_t) * count))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate object method descriptor list)))
    /* Deallocate object interface. */
    free(*interface);
    *interface = NULL;
    errno = errno2;
    return -1;
  }
  /* Duplicate object method descriptor list.*/
  for (index = 0; index < count; index++) {
    /* Duplicate object method name. */
    if (NULL == ((*interface)->methods[index].name =
                 strdup(methods[index].name))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
      /* Deallocate object method descriptor list.*/
      for (index = 0; index < (*interface)->count; index++) {
        /* Deallocate object method name. */
        free((*interface)->methods[index].name);
        /* Deallocate object method output signature. */
        free((*interface)->methods[index].output);
        /* Deallocate object method input signature. */
        free((*interface)->methods[index].input);
      }
      /* Deallocate object method descriptor list. */
      free((*interface)->methods);
      /* Deallocate object interface. */
      free(*interface);
      *interface = NULL;
      errno = errno2;
      return -1;
    }
    /* Duplicate object method output signature. */
    if (NULL == ((*interface)->methods[index].output =
                 strdup(methods[index].output))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
      /* Deallocate object method name. */
      free((*interface)->methods[index].name);
      /* Deallocate object method descriptor list.*/
      for (index = 0; index < (*interface)->count; index++) {
        /* Deallocate object method name. */
        free((*interface)->methods[index].name);
        /* Deallocate object method output signature. */
        free((*interface)->methods[index].output);
        /* Deallocate object method input signature. */
        free((*interface)->methods[index].input);
      }
      /* Deallocate object method descriptor list. */
      free((*interface)->methods);
      /* Deallocate object interface. */
      free(*interface);
      *interface = NULL;
      errno = errno2;
      return -1;
    }
    /* Duplicate object method input signature. */
    if (NULL == ((*interface)->methods[index].input =
                 strdup(methods[index].input))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
      /* Deallocate object method name. */
      free((*interface)->methods[index].name);
      /* Deallocate object method output signature. */
      free((*interface)->methods[index].output);
      /* Deallocate object method descriptor list.*/
      for (index = 0; index < (*interface)->count; index++) {
        /* Deallocate object method name. */
        free((*interface)->methods[index].name);
        /* Deallocate object method output signature. */
        free((*interface)->methods[index].output);
        /* Deallocate object method input signature. */
        free((*interface)->methods[index].input);
      }
      /* Deallocate object method descriptor list. */
      free((*interface)->methods);
      /* Deallocate object interface. */
      free(*interface);
      *interface = NULL;
      errno = errno2;
      return -1;
    }
    /* Copy object method function. */
    (*interface)->methods[index].call = methods[index].call;
    /* Increase object method descriptor list count. */
    (*interface)->count++;
  }
  return 0;
}

/*
 * Destroys an object interface.
 *
 * interface = The object interface to destroy (also return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_destroy"
int rmix_interface_destroy (rmix_interface_t **interface) {
  unsigned int index;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check *interface parameter. */
  if (NULL == *interface) {
    return 0;
  }
  /* Deallocate object method descriptor list entries. */
  for (index = 0; index < (*interface)->count; index++) {
    /* Deallocate object method name. */
    free((*interface)->methods[index].name);
    /* Deallocate object method output signature. */
    free((*interface)->methods[index].output);
    /* Deallocate object method input signature. */
    free((*interface)->methods[index].input);
  }
  /* Deallocate object method descriptor list. */
  if (NULL != (*interface)->methods) {
    free((*interface)->methods);
  }
  /* Deallocate object interface. */
  free(*interface);
  *interface = NULL;
  return 0;
}

/*
 * Duplicates an object interface.
 *
 * interface1 = The object interface to create (return).
 * interface2 = The object interface to duplicate.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_duplicate"
int rmix_interface_duplicate (rmix_interface_t       **interface1,
                              const rmix_interface_t  *interface2) {
#ifdef DEBUG
  /* Check interface1 parameter. */
  if (NULL == interface1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2 parameter. */
  if (NULL == interface2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create object interface. */
  if (0 != rmix_interface_create2(interface1, interface2->count,
                                  interface2->methods)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create object interface)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Appends an object interface to another object interface.
 *
 * interface1 = The object interface.
 * interface2 = The object interface to append.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_append"
int rmix_interface_append (rmix_interface_t       *interface1,
                           const rmix_interface_t *interface2) {
  unsigned int   index1, index2; 
  rmix_method_t *methods;
#ifdef DEBUG
  /* Check interface1 parameter. */
  if (NULL == interface1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->methods parameter. */
  if ((NULL == interface1->methods)&&(0 != interface1->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->count parameter. */
  if ((0 == interface1->count)&&(NULL != interface1->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2 parameter. */
  if (NULL == interface2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->methods parameter. */
  if ((NULL == interface2->methods)&&(0 != interface2->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->count parameter. */
  if ((0 == interface2->count)&&(NULL != interface2->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Rellocate object method descriptor list. */
  if (NULL == (methods =
               realloc(interface1->methods, sizeof(rmix_method_t) *
                       (interface1->count + interface2->count)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object method descriptor list)))
    errno = errno2;
    return -1;
  }
  interface1->methods = methods;
  /* Duplicate object method descriptor list.*/
  for (index1 = interface1->count, index2 = 0;
       index2 < interface2->count; index1++, index2++) {
    /* Duplicate object method name. */
    if (NULL == (interface1->methods[index1].name =
                 strdup(interface2->methods[index2].name))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
      /* Deallocate object method descriptor list.*/
      for (index1 = interface1->count; 0 != index2; index1++, index2--) {
        /* Deallocate object method name. */
        free(interface1->methods[index1].name);
        /* Deallocate object method output signature. */
        free(interface1->methods[index1].output);
        /* Deallocate object method input signature. */
        free(interface1->methods[index1].input);
      }
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface1->methods,
                                        sizeof(rmix_method_t) *
                                        interface1->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
      } else {
        interface1->methods = methods;
      }
      errno = errno2;
      return -1;
    }
    /* Duplicate object method output signature. */
    if (NULL == (interface1->methods[index1].output =
                 strdup(interface2->methods[index2].output))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
      /* Deallocate object method name. */
      free(interface1->methods[index1].name);
      /* Deallocate object method descriptor list.*/
      for (index1 = interface1->count; 0 != index2; index1++, index2--) {
        /* Deallocate object method name. */
        free(interface1->methods[index1].name);
        /* Deallocate object method output signature. */
        free(interface1->methods[index1].output);
        /* Deallocate object method input signature. */
        free(interface1->methods[index1].input);
      }
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface1->methods,
                                        sizeof(rmix_method_t) *
                                        interface1->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
      } else {
        interface1->methods = methods;
      }
      errno = errno2;
      return -1;
    }
    /* Duplicate object method input signature. */
    if (NULL == (interface1->methods[index1].input =
                 strdup(interface2->methods[index2].input))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
      /* Deallocate object method name. */
      free(interface1->methods[index1].name);
      /* Deallocate object method output signature. */
      free(interface1->methods[index1].output);
      /* Deallocate object method descriptor list.*/
      for (index1 = interface1->count; 0 != index2; index1++, index2--) {
        /* Deallocate object method name. */
        free(interface1->methods[index1].name);
        /* Deallocate object method output signature. */
        free(interface1->methods[index1].output);
        /* Deallocate object method input signature. */
        free(interface1->methods[index1].input);
      }
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface1->methods,
                                        sizeof(rmix_method_t) *
                                        interface1->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
      } else {
        interface1->methods = methods;
      }
      errno = errno2;
      return -1;
    }
    /* Copy object method function. */
    interface1->methods[index1].call = interface2->methods[index2].call;
  }
  /* Increase object method descriptor list count. */
  interface1->count += interface2->count;
  return 0;
}

/*
 * Inserts an object interface to another object interface.
 *
 * interface1 = The object interface.
 * index      = The object method descriptor list index to insert.
 * interface2 = The object interface to insert.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_insert"
int rmix_interface_insert (rmix_interface_t       *interface1,
                           unsigned int            index,
                           const rmix_interface_t *interface2) {
  unsigned int   index1, index2;
  rmix_method_t *methods;
#ifdef DEBUG
  /* Check interface1 parameter. */
  if (NULL == interface1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->methods parameter. */
  if ((NULL == interface1->methods)&&(0 != interface1->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->count parameter. */
  if ((0 == interface1->count)&&(NULL != interface1->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface1->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2 parameter. */
  if (NULL == interface2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->methods parameter. */
  if ((NULL == interface2->methods)&&(0 != interface2->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->count parameter. */
  if ((0 == interface2->count)&&(NULL != interface2->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Rellocate object method descriptor list. */
  if (NULL == (methods =
               realloc(interface1->methods, sizeof(rmix_method_t) *
                       (interface1->count + interface2->count)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object method descriptor list)))
    errno = errno2;
    return -1;
  }
  interface1->methods = methods;
  memmove(interface1->methods + index + interface2->count,
          interface1->methods + index,
          (interface1->count - index) * sizeof(rmix_method_t));
  /* Duplicate object method descriptor list.*/
  for (index1 = index, index2 = 0;
       index2 < interface2->count; index1++, index2++) {
    /* Duplicate object method name. */
    if (NULL == (interface1->methods[index1].name =
                 strdup(interface2->methods[index2].name))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
      /* Deallocate object method descriptor list.*/
      for (index1 = index; 0 != index2; index1++, index2--) {
        /* Deallocate object method name. */
        free(interface1->methods[index1].name);
        /* Deallocate object method output signature. */
        free(interface1->methods[index1].output);
        /* Deallocate object method input signature. */
        free(interface1->methods[index1].input);
      }
      memmove(interface1->methods + index,
              interface1->methods + index + interface2->count,
             (interface1->count - index) * sizeof(rmix_method_t));
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface1->methods,
                                        sizeof(rmix_method_t) *
                                        interface1->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
      } else {
        interface1->methods = methods;
      }
      errno = errno2;
      return -1;
    }
    /* Duplicate object method output signature. */
    if (NULL == (interface1->methods[index1].output =
                 strdup(interface2->methods[index2].output))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
      /* Deallocate object method name. */
      free(interface1->methods[index1].name);
      /* Deallocate object method descriptor list.*/
      for (index1 = index; 0 != index2; index1++, index2--) {
        /* Deallocate object method name. */
        free(interface1->methods[index1].name);
        /* Deallocate object method output signature. */
        free(interface1->methods[index1].output);
        /* Deallocate object method input signature. */
        free(interface1->methods[index1].input);
      }
      memmove(interface1->methods + index,
              interface1->methods + index + interface2->count,
             (interface1->count - index) * sizeof(rmix_method_t));
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface1->methods,
                                        sizeof(rmix_method_t) *
                                        interface1->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
      } else {
        interface1->methods = methods;
      }
      errno = errno2;
      return -1;
    }
    /* Duplicate object method input signature. */
    if (NULL == (interface1->methods[index1].input =
                 strdup(interface2->methods[index2].input))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
      /* Deallocate object method name. */
      free(interface1->methods[index1].name);
      /* Deallocate object method output signature. */
      free(interface1->methods[index1].output);
      /* Deallocate object method descriptor list.*/
      for (index1 = index; 0 != index2; index1++, index2--) {
        /* Deallocate object method name. */
        free(interface1->methods[index1].name);
        /* Deallocate object method output signature. */
        free(interface1->methods[index1].output);
        /* Deallocate object method input signature. */
        free(interface1->methods[index1].input);
      }
      memmove(interface1->methods + index,
              interface1->methods + index + interface2->count,
             (interface1->count - index) * sizeof(rmix_method_t));
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface1->methods,
                                        sizeof(rmix_method_t) *
                                        interface1->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
      } else {
        interface1->methods = methods;
      }
      errno = errno2;
      return -1;
    }
    /* Copy object method function. */
    interface1->methods[index1].call = interface2->methods[index2].call;
  }
  /* Increase object method descriptor list count. */
  interface1->count += interface2->count;
  return 0;
}

/*
 * Removes an object interface from another object interface.
 *
 * interface1 = The object interface.
 * interface2 = The object interface to remove.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_remove"
int rmix_interface_remove (rmix_interface_t       *interface1,
                           const rmix_interface_t *interface2) {
  unsigned int   index, index1, index2;
  rmix_method_t *methods;
#ifdef DEBUG
  /* Check interface1 parameter. */
  if (NULL == interface1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->methods parameter. */
  if ((NULL == interface1->methods)&&(0 != interface1->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->count parameter. */
  if ((0 == interface1->count)&&(NULL != interface1->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2 parameter. */
  if (NULL == interface2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->methods parameter. */
  if ((NULL == interface2->methods)&&(0 != interface2->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->count parameter. */
  if ((0 == interface2->count)&&(NULL != interface2->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for object interface2 in object interface1. */
  for (index = 0; index < interface1->count; index++) {
    for (index1 = index, index2 = 0;
         (index1 < interface1->count)&&(index2 < interface2->count)&&
         (0 == strcmp(interface1->methods[index1].name,
                      interface2->methods[index2].name  ))&&
         (0 == strcmp(interface1->methods[index1].input,
                      interface2->methods[index2].input ))&&
         (0 == strcmp(interface1->methods[index1].output,
                      interface2->methods[index2].output))&&
          (interface1->methods[index1].call ==
           interface2->methods[index2].call); index1++, index2++);
    /* Check for object interface2 in object interface1. */
    if (index2 == interface2->count) {
      /* Deallocate object method descriptor list.*/
      for (index1 = index, index2 = index1 + interface2->count;
           index1 < index2; index1++) {
        /* Deallocate object method name. */
        free(interface1->methods[index1].name);
        /* Deallocate object method output signature. */
        free(interface1->methods[index1].output);
        /* Deallocate object method input signature. */
        free(interface1->methods[index1].input);
      }
      interface1->count -= interface2->count;
      memmove(interface1->methods + index,
              interface1->methods + index + interface2->count,
             (interface1->count - index) * sizeof(rmix_method_t));
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface1->methods,
                                        sizeof(rmix_method_t) *
                                        interface1->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
        errno = 0;
      } else {
        interface1->methods = methods;
      }
      return 0;
    }
  }
  errno = ESRCH;
  return -1;
}

/*
 * Finds an object interface in another object interface.
 *
 * interface1 = The object interface.
 * index      = The object method descriptor list index (return).
 * interface2 = The object interface to find.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_find"
int rmix_interface_find (const rmix_interface_t *interface1,
                         unsigned int           *index,
                         const rmix_interface_t *interface2) {
  unsigned int index1, index2;
#ifdef DEBUG
  /* Check interface1 parameter. */
  if (NULL == interface1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->methods parameter. */
  if ((NULL == interface1->methods)&&(0 != interface1->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->count parameter. */
  if ((0 == interface1->count)&&(NULL != interface1->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (NULL == index) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2 parameter. */
  if (NULL == interface2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->methods parameter. */
  if ((NULL == interface2->methods)&&(0 != interface2->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->count parameter. */
  if ((0 == interface2->count)&&(NULL != interface2->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for object interface2 in object interface1. */
  for (*index = 0; *index < interface1->count; (*index)++) {
    for (index1 = *index, index2 = 0;
         (index1 < interface1->count)&&(index2 < interface2->count)&&
         (0 == strcmp(interface1->methods[index1].name,
                      interface2->methods[index2].name  ))&&
         (0 == strcmp(interface1->methods[index1].input,
                      interface2->methods[index2].input ))&&
         (0 == strcmp(interface1->methods[index1].output,
                      interface2->methods[index2].output))&&
          (interface1->methods[index1].call ==
           interface2->methods[index2].call); index1++, index2++);
    /* Check for object interface2 in object interface1. */
    if (index2 == interface2->count) {
      return 0;
    }
  }
  errno = ESRCH;
  return -1;
}

/*
 * Splits an object interface in two object interfaces.
 *
 * interface = The object interface.
 * head      = The object interface head part.
 * tail      = The object interface tail part.
 * index     = The object method descriptor list index to split.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_split"
int rmix_interface_split (const rmix_interface_t  *interface,
                          rmix_interface_t       **head,
                          rmix_interface_t       **tail,
                          unsigned int             index) {
  unsigned int index1;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check head parameter. */
  if (NULL == head) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(head parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check tail parameter. */
  if (NULL == tail) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(tail parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate head object interface. */
  if (NULL == (*head = malloc(sizeof(rmix_interface_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate head object interface)))
    errno = errno2;
    return -1;
  }
  (*head)->count = 0;
  /* Allocate head object method descriptor list. */
  if (NULL == ((*head)->methods =
               malloc(sizeof(rmix_method_t) * interface->count))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate head object method descriptor list)))
    /* Deallocate object interface. */
    free(*head);
    *head = NULL;
    errno = errno2;
    return -1;
  }
  /* Duplicate head object method descriptor list.*/
  for (index1 = 0; index1 < index; index1++) {
    /* Duplicate head object method name. */
    if (NULL == ((*head)->methods[index1].name =
                 strdup(interface->methods[index1].name))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to duplicate head object method name)))
      /* Deallocate head object method descriptor list.*/
      for (index1 = 0; index1 < (*head)->count; index1++) {
        /* Deallocate head object method name. */
        free((*head)->methods[index1].name);
        /* Deallocate head object method output signature. */
        free((*head)->methods[index1].output);
        /* Deallocate head object method input signature. */
        free((*head)->methods[index1].input);
      }
      /* Deallocate head object method descriptor list. */
      free((*head)->methods);
      /* Deallocate head object interface. */
      free(*head);
      *head = NULL;
      errno = errno2;
      return -1;
    }
    /* Duplicate head object method output signature. */
    if (NULL == ((*head)->methods[index1].output =
                 strdup(interface->methods[index1].output))) {
      int errno2 = errno;
      RMIX_LOG((
        RMIX_WARN(unable to duplicate head object method output signature)))
      /* Deallocate head object method name. */
      free((*head)->methods[index1].name);
      /* Deallocate head object method descriptor list.*/
      for (index1 = 0; index1 < (*head)->count; index1++) {
        /* Deallocate head object method name. */
        free((*head)->methods[index1].name);
        /* Deallocate head object method output signature. */
        free((*head)->methods[index1].output);
        /* Deallocate head object method input signature. */
        free((*head)->methods[index1].input);
      }
      /* Deallocate head object method descriptor list. */
      free((*head)->methods);
      /* Deallocate head object interface. */
      free(*head);
      *head = NULL;
      errno = errno2;
      return -1;
    }
    /* Duplicate head object method input signature. */
    if (NULL == ((*head)->methods[index1].input =
                 strdup(interface->methods[index1].input))) {
      int errno2 = errno;
      RMIX_LOG((
        RMIX_WARN(unable to duplicate head object method input signature)))
      /* Deallocate head object method name. */
      free((*head)->methods[index1].name);
      /* Deallocate head object method output signature. */
      free((*head)->methods[index1].output);
      /* Deallocate head object method descriptor list.*/
      for (index1 = 0; index1 < (*head)->count; index1++) {
        /* Deallocate head object method name. */
        free((*head)->methods[index1].name);
        /* Deallocate head object method output signature. */
        free((*head)->methods[index1].output);
        /* Deallocate head object method input signature. */
        free((*head)->methods[index1].input);
      }
      /* Deallocate head object method descriptor list. */
      free((*head)->methods);
      /* Deallocate head object interface. */
      free(*head);
      *head = NULL;
      errno = errno2;
      return -1;
    }
    /* Copy head object method function. */
    (*head)->methods[index1].call = interface->methods[index1].call;
    /* Increase head object method descriptor list count. */
    (*head)->count++;
  }
  /* Allocate tail object interface. */
  if (NULL == (*tail = malloc(sizeof(rmix_interface_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate tail object interface)))
    /* Deallocate head object method descriptor list.*/
    for (index1 = 0; index1 < (*head)->count; index1++) {
      /* Deallocate head object method name. */
      free((*head)->methods[index1].name);
      /* Deallocate head object method output signature. */
      free((*head)->methods[index1].output);
      /* Deallocate head object method input signature. */
      free((*head)->methods[index1].input);
    }
    /* Deallocate head object method descriptor list. */
    free((*head)->methods);
    /* Deallocate head object interface. */
    free(*head);
    errno = errno2;
    return -1;
  }
  (*tail)->count = 0;
  /* Allocate tail object method descriptor list. */
  if (NULL == ((*tail)->methods =
               malloc(sizeof(rmix_method_t) * interface->count))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate tail object method descriptor list)))
    /* Deallocate object interface. */
    free(*tail);
    *tail = NULL;
    /* Deallocate head object method descriptor list.*/
    for (index1 = 0; index1 < (*head)->count; index1++) {
      /* Deallocate head object method name. */
      free((*head)->methods[index1].name);
      /* Deallocate head object method output signature. */
      free((*head)->methods[index1].output);
      /* Deallocate head object method input signature. */
      free((*head)->methods[index1].input);
    }
    /* Deallocate head object method descriptor list. */
    free((*head)->methods);
    /* Deallocate head object interface. */
    free(*head);
    errno = errno2;
    return -1;
  }
  /* Duplicate tail object method descriptor list.*/
  for (index1 = index; index1 < interface->count; index1++) {
    /* Duplicate tail object method name. */
    if (NULL == ((*tail)->methods[index1].name =
                 strdup(interface->methods[index1].name))) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to tail duplicate object method name)))
      /* Deallocate tail object method descriptor list.*/
      for (index1 = 0; index1 < (*tail)->count; index1++) {
        /* Deallocate tail object method name. */
        free((*tail)->methods[index1].name);
        /* Deallocate tail object method output signature. */
        free((*tail)->methods[index1].output);
        /* Deallocate tail object method input signature. */
        free((*tail)->methods[index1].input);
      }
      /* Deallocate tail object method descriptor list. */
      free((*tail)->methods);
      /* Deallocate tail object interface. */
      free(*tail);
      *tail = NULL;
      /* Deallocate head object method descriptor list.*/
      for (index1 = 0; index1 < (*head)->count; index1++) {
        /* Deallocate head object method name. */
        free((*head)->methods[index1].name);
        /* Deallocate head object method output signature. */
        free((*head)->methods[index1].output);
        /* Deallocate head object method input signature. */
        free((*head)->methods[index1].input);
      }
      /* Deallocate head object method descriptor list. */
      free((*head)->methods);
      /* Deallocate head object interface. */
      free(*head);
      *head = NULL;
      errno = errno2;
      return -1;
    }
    /* Duplicate tail object method output signature. */
    if (NULL == ((*tail)->methods[index1].output =
                 strdup(interface->methods[index1].output))) {
      int errno2 = errno;
      RMIX_LOG((
        RMIX_WARN(unable to duplicate tail object method output signature)))
      /* Deallocate tail object method name. */
      free((*tail)->methods[index1].name);
      /* Deallocate tail object method descriptor list.*/
      for (index1 = 0; index1 < (*tail)->count; index1++) {
        /* Deallocate tail object method name. */
        free((*tail)->methods[index1].name);
        /* Deallocate tail object method output signature. */
        free((*tail)->methods[index1].output);
        /* Deallocate tail object method input signature. */
        free((*tail)->methods[index1].input);
      }
      /* Deallocate tail object method descriptor list. */
      free((*tail)->methods);
      /* Deallocate tail object interface. */
      free(*tail);
      *tail = NULL;
      /* Deallocate head object method descriptor list.*/
      for (index1 = 0; index1 < (*head)->count; index1++) {
        /* Deallocate head object method name. */
        free((*head)->methods[index1].name);
        /* Deallocate head object method output signature. */
        free((*head)->methods[index1].output);
        /* Deallocate head object method input signature. */
        free((*head)->methods[index1].input);
      }
      /* Deallocate head object method descriptor list. */
      free((*head)->methods);
      /* Deallocate head object interface. */
      free(*head);
      *head = NULL;
      errno = errno2;
      return -1;
    }
    /* Duplicate tail object method input signature. */
    if (NULL == ((*tail)->methods[index1].input =
                 strdup(interface->methods[index1].input))) {
      int errno2 = errno;
      RMIX_LOG((
        RMIX_WARN(unable to duplicate tail object method input signature)))
      /* Deallocate tail object method name. */
      free((*tail)->methods[index1].name);
      /* Deallocate tail object method output signature. */
      free((*tail)->methods[index1].output);
      /* Deallocate tail object method descriptor list.*/
      for (index1 = 0; index1 < (*tail)->count; index1++) {
        /* Deallocate tail object method name. */
        free((*tail)->methods[index1].name);
        /* Deallocate tail object method output signature. */
        free((*tail)->methods[index1].output);
        /* Deallocate tail object method input signature. */
        free((*tail)->methods[index1].input);
      }
      /* Deallocate tail object method descriptor list. */
      free((*tail)->methods);
      /* Deallocate tail object interface. */
      free(*tail);
      *tail = NULL;
      /* Deallocate head object method descriptor list.*/
      for (index1 = 0; index1 < (*head)->count; index1++) {
        /* Deallocate head object method name. */
        free((*head)->methods[index1].name);
        /* Deallocate head object method output signature. */
        free((*head)->methods[index1].output);
        /* Deallocate head object method input signature. */
        free((*head)->methods[index1].input);
      }
      /* Deallocate head object method descriptor list. */
      free((*head)->methods);
      /* Deallocate head object interface. */
      free(*head);
      *head = NULL;
      errno = errno2;
      return -1;
    }
    /* Copy tail object method function. */
    (*tail)->methods[index1].call = interface->methods[index1].call;
    /* Increase tail object method descriptor list count. */
    (*tail)->count++;
  }
  return 0;
}

/*
 * Compares two object interfaces.
 *
 * diff       = The difference between both object interfaces (return).
 *              diff > 0 if interface1 < interface2,
 *              diff < 0 if interface1 > interface2, and
 *              diff = 0 if interface1 = interface2
 * interface1 = The object interface to compare against.
 * interface2 = The object interface to compare with.
 * return     = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_compare"
int rmix_interface_compare (int                    *diff,
                            const rmix_interface_t *interface1,
                            const rmix_interface_t *interface2) {
  unsigned int index;
#ifdef DEBUG
  /* Check diff parameter. */
  if (NULL == diff) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(diff parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1 parameter. */
  if (NULL == interface1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->methods parameter. */
  if ((NULL == interface1->methods)&&(0 != interface1->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface1->count parameter. */
  if ((0 == interface1->count)&&(NULL != interface1->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface1->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2 parameter. */
  if (NULL == interface2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->methods parameter. */
  if ((NULL == interface2->methods)&&(0 != interface2->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface2->count parameter. */
  if ((0 == interface2->count)&&(NULL != interface2->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface2->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Compare and check for equal signature list size. */
  if (0 == (*diff = interface2->count - interface1->count)) {
    /* Compare and check for equal methods. */
    for (index = 0; index < interface1->count; index++) {
      /* Compare and check for equal merthod name, output
         signature, input signature and method function. */
      if ((0 != (*diff = strcmp(interface1->methods[index].name,
                                interface2->methods[index].name)))||
          (0 != (*diff = strcmp(interface1->methods[index].output,
                                interface2->methods[index].output)))||
          (0 != (*diff = strcmp(interface1->methods[index].input,
                                interface2->methods[index].input)))||
          (0 != (*diff = interface2->methods[index].call -
                         interface1->methods[index].call))) {
        return 0;
      }
    }
  }
  return 0;
}

/*
 * Appends an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * name      = The object method name to append.
 * output    = The object method output signature to append.
 * input     = The object method input  signature to append.
 * method    = The object method function to append.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_append"
int rmix_interface_methods_append (rmix_interface_t   *interface,
                                   const char         *name,
                                   const char         *output,
                                   const char         *input,
                                   rmix_method_call_t  call) {
  rmix_method_t *methods;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check output parameter. */
  if (NULL == output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check input parameter. */
  if (NULL == input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check call parameter. */
  if (NULL == call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(call parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Rellocate object method descriptor list. */
  if (NULL == (methods = realloc(interface->methods,
                                 sizeof(rmix_method_t) *
                                 (interface->count + 1)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object method descriptor list)))
    errno = errno2;
    return -1;
  }
  interface->methods = methods;
  /* Duplicate object method name. */
  if (NULL == (interface->methods[interface->count].name = strdup(name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method output signature. */
  if (NULL == (interface->methods[interface->count].output =
               strdup(output))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
    /* Deallocate object method name. */
    free(interface->methods[interface->count].name);
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method input signature. */
  if (NULL == (interface->methods[interface->count].input = strdup(input))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
    /* Deallocate object method name. */
    free(interface->methods[interface->count].name);
    /* Deallocate object method output signature. */
    free(interface->methods[interface->count].output);
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Copy object method function. */
  interface->methods[interface->count].call = call;
  /* Increase object method descriptor list size. */
  interface->count++;
  return 0;
}

/*
 * Appends an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * method    = The object method descriptor to append.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_append2"
int rmix_interface_methods_append2 (rmix_interface_t    *interface,
                                    const rmix_method_t *method) {
  rmix_method_t *methods;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check method parameter. */
  if (NULL == method) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->name parameter. */
  if (NULL == method->name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->output parameter. */
  if (NULL == method->output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->input parameter. */
  if (NULL == method->input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->call parameter. */
  if (NULL == method->call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->call parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Rellocate object method descriptor list. */
  if (NULL == (methods = realloc(interface->methods,
                                 sizeof(rmix_method_t) *
                                 (interface->count + 1)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object method descriptor list)))
    errno = errno2;
    return -1;
  }
  interface->methods = methods;
  /* Duplicate object method name. */
  if (NULL == (interface->methods[interface->count].name =
               strdup(method->name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method output signature. */
  if (NULL == (interface->methods[interface->count].output =
               strdup(method->output))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
    /* Deallocate object method name. */
    free(interface->methods[interface->count].name);
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method input signature. */
  if (NULL == (interface->methods[interface->count].input =
               strdup(method->input))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
    /* Deallocate object method name. */
    free(interface->methods[interface->count].name);
    /* Deallocate object method output signature. */
    free(interface->methods[interface->count].output);
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Copy object method function. */
  interface->methods[interface->count].call = method->call;
  /* Increase object method descriptor list size. */
  interface->count++;
  return 0;
}

/*
 * Inserts an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index to insert.
 * name      = The object method name to insert.
 * output    = The object method output signature to insert.
 * input     = The object method input  signature to insert.
 * method    = The object method function to insert.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_insert"
int rmix_interface_methods_insert (rmix_interface_t   *interface,
                                   unsigned int        index,
                                   const char         *name,
                                   const char         *output,
                                   const char         *input,
                                   rmix_method_call_t  call) {
  rmix_method_t *methods;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check output parameter. */
  if (NULL == output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check input parameter. */
  if (NULL == input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check call parameter. */
  if (NULL == call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(call parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Rellocate object method descriptor list. */
  if (NULL == (methods = realloc(interface->methods,
                                 sizeof(rmix_method_t) *
                                 (interface->count + 1)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object method descriptor list)))
    errno = errno2;
    return -1;
  }
  interface->methods = methods;
  memmove(interface->methods + index + 1,
          interface->methods + index,
          (interface->count - index) * sizeof(rmix_method_t));
  /* Duplicate object method name. */
  if (NULL == (interface->methods[index].name = strdup(name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
    memmove(interface->methods + index,
            interface->methods + index + 1,
           (interface->count - index) * sizeof(rmix_method_t));
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method output signature. */
  if (NULL == (interface->methods[index].output = strdup(output))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
    /* Deallocate object method name. */
    free(interface->methods[index].name);
    memmove(interface->methods + index,
            interface->methods + index + 1,
           (interface->count - index) * sizeof(rmix_method_t));
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((
        RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method input signature. */
  if (NULL == (interface->methods[index].input = strdup(input))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
    /* Deallocate object method name. */
    free(interface->methods[index].name);
    /* Deallocate object method output signature. */
    free(interface->methods[index].output);
    memmove(interface->methods + index,
            interface->methods + index + 1,
           (interface->count - index) * sizeof(rmix_method_t));
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Copy object method function. */
  interface->methods[interface->count].call = call;
  /* Increase object method descriptor list size. */
  interface->count++;
  return 0;
}

/*
 * Inserts an object method descriptor to the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index to insert.
 * method    = The object method descriptor to insert.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_insert2"
int rmix_interface_methods_insert2 (rmix_interface_t    *interface,
                                    unsigned int         index,
                                    const rmix_method_t *method) {
  rmix_method_t *methods;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
  /* Check method parameter. */
  if (NULL == method) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->name parameter. */
  if (NULL == method->name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->output parameter. */
  if (NULL == method->output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->input parameter. */
  if (NULL == method->input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->call parameter. */
  if (NULL == method->call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->call parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Rellocate object method descriptor list. */
  if (NULL == (methods = realloc(interface->methods,
                                 sizeof(rmix_method_t) *
                                 (interface->count + 1)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object method descriptor list)))
    errno = errno2;
    return -1;
  }
  interface->methods = methods;
  memmove(interface->methods + index + 1,
          interface->methods + index,
          (interface->count - index) * sizeof(rmix_method_t));
  /* Duplicate object method name. */
  if (NULL == (interface->methods[index].name = strdup(method->name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
    /* Decrease object method descriptor list. */
    memmove(interface->methods + index,
            interface->methods + index + 1,
           (interface->count - index) * sizeof(rmix_method_t));
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method output signature. */
  if (NULL == (interface->methods[index].output = strdup(method->output))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
    /* Deallocate object method name. */
    free(interface->methods[index].name);
    /* Decrease object method descriptor list. */
    memmove(interface->methods + index,
            interface->methods + index + 1,
           (interface->count - index) * sizeof(rmix_method_t));
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Duplicate object method input signature. */
  if (NULL == (interface->methods[index].input = strdup(method->input))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
    /* Deallocate object method name. */
    free(interface->methods[index].name);
    /* Deallocate object method output signature. */
    free(interface->methods[index].output);
    /* Decrease object method descriptor list. */
    memmove(interface->methods + index,
            interface->methods + index + 1,
           (interface->count - index) * sizeof(rmix_method_t));
    /* Rellocate object method descriptor list. */
    if (NULL == (methods = realloc(interface->methods,
                                   sizeof(rmix_method_t) *
                                   interface->count))) {
      RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    } else {
      interface->methods = methods;
    }
    errno = errno2;
    return -1;
  }
  /* Copy object method function. */
  interface->methods[interface->count].call = method->call;
  /* Increase object method descriptor list size. */
  interface->count++;
  return 0;
}

/*
 * Removes an object method descriptor from the object interface.
 *
 * interface = The object interface.
 * name      = The object method name to remove.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_remove"
int rmix_interface_methods_remove (rmix_interface_t *interface,
                                   const char       *name) {
  unsigned int   index;
  rmix_method_t *methods;
 #ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for object method name. */
  for (index = 0; index < interface->count; index++) {
    /* Check for object method name. */
    if (0 == strcmp(interface->methods[index].name, name)) {
      /* Deallocate object method name. */
      free(interface->methods[index].name);
      /* Deallocate object method output signature. */
      free(interface->methods[index].output);
      /* Deallocate object method input signature. */
      free(interface->methods[index].input);
      /* Decrease object method descriptor list. */
      interface->count--;
      memmove(interface->methods + index,
              interface->methods + index + 1,
             (interface->count - index) * sizeof(rmix_method_t));
      /* Rellocate object method descriptor list. */
      if (NULL == (methods = realloc(interface->methods,
                                        sizeof(rmix_method_t) *
                                        interface->count))) {
        RMIX_LOG((
          RMIX_INFO(unable to reallocate object method descriptor list)))
        errno = 0;
      } else {
        interface->methods = methods;
      }
      return 0;
    }
  }
  errno = ESRCH;
  return -1;
}

/*
 * Removes an object method descriptor from the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index to remove.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_remove2"
int rmix_interface_methods_remove2 (rmix_interface_t *interface,
                                    unsigned int      index) {
  rmix_method_t *methods;
 #ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Deallocate object method name. */
  free(interface->methods[index].name);
  /* Deallocate object method output signature. */
  free(interface->methods[index].output);
  /* Deallocate object method input signature. */
  free(interface->methods[index].input);
  /* Decrease object method descriptor list. */
  interface->count--;
  memmove(interface->methods + index,
          interface->methods + index + 1,
         (interface->count - index) * sizeof(rmix_method_t));
  /* Rellocate object method descriptor list. */
  if (NULL == (methods = realloc(interface->methods,
                                    sizeof(rmix_method_t) *
                                    interface->count))) {
    RMIX_LOG((RMIX_INFO(unable to reallocate object method descriptor list)))
    errno = 0;
  } else {
    interface->methods = methods;
  }
  return 0;
}

/*
 * Finds an object method descriptor in the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index (return).
 * name      = The object method name to find.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_find"
int rmix_interface_methods_find (const rmix_interface_t *interface,
                                 unsigned int           *index,
                                 const char             *name) {
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (NULL == index) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for object method name. */
  for (*index = 0; *index < interface->count; (*index)++) {
    /* Check for object method name. */
    if (0 == strcmp(interface->methods[*index].name, name)) {
      return 0;
    }
  }
  errno = ESRCH;
  return -1;
}

/*
 * Finds an object method descriptor in the object interface.
 *
 * interface = The object interface.
 * index     = The object method descriptor list index (return).
 * method    = The object method descriptor to find.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_find2"
int rmix_interface_methods_find2 (const rmix_interface_t *interface,
                                  unsigned int           *index,
                                  const rmix_method_t    *method) {
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (NULL == index) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method parameter. */
  if (NULL == method) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->name parameter. */
  if (NULL == method->name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->output parameter. */
  if (NULL == method->output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->input parameter. */
  if (NULL == method->input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method->call parameter. */
  if (NULL == method->call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method->call parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Search for object method descriptor. */
  for (*index = 0; *index < interface->count; (*index)++) {
    /* Check for object method descriptor. */
    if ((0 == strcmp(interface->methods[*index].name,   method->name  ))&&
        (0 == strcmp(interface->methods[*index].input,  method->input ))&&
        (0 == strcmp(interface->methods[*index].output, method->output))&&
        (interface->methods[*index].call == method->call)) {
      return 0;
    }
  }
  errno = ESRCH;
  return -1;
}

/*
 * Gets the object method name of the object interface.
 *
 * interface = The object interface.
 * name      = The object method name (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_name"
int rmix_interface_methods_name (const rmix_interface_t  *interface,
                                 const char             **name,
                                 unsigned int             index) {
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  *name = interface->methods[index].name;
  return 0;
}

/*
 * Gets the object method output signature of the object interface.
 *
 * interface = The object interface.
 * output    = The object method output signature (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_output"
int rmix_interface_methods_output (const rmix_interface_t  *interface,
                                   const char             **output,
                                   unsigned int             index) {
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check output parameter. */
  if (NULL == output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  *output = interface->methods[index].output;
  return 0;
}

/*
 * Gets the object method input signature of the object interface.
 *
 * interface = The object interface.
 * input     = The object method input signature (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_input"
int rmix_interface_methods_input (const rmix_interface_t  *interface,
                                  const char             **input,
                                  unsigned int             index) {
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check input parameter. */
  if (NULL == input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  *input = interface->methods[index].input;
  return 0;
}

/*
 * Gets the object method function of the object interface.
 *
 * interface = The object interface.
 * method    = The object method function (return).
 * index     = The object method descriptor list index.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_interface_methods_call"
int rmix_interface_methods_call (const rmix_interface_t *interface,
                                 rmix_method_call_t     *call,
                                 unsigned int            index) {
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->methods parameter. */
  if ((NULL == interface->methods)&&(0 != interface->count)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->methods parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface->count parameter. */
  if ((0 == interface->count)&&(NULL != interface->methods)) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface->count parameter is zero)))
    errno = EINVAL;
    return -1;
  }
  /* Check call parameter. */
  if (NULL == call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(call parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check index parameter. */
  if (index > interface->count) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(index parameter is too big)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  *call = interface->methods[index].call;
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
