/******************************************************************************
 *
 * Header file for rmix-rpcx auth (authentication) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_RPCX_AUTH_H
#define RMIX_RPCX_AUTH_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates an authentication credentials/verifier pair.
 *
 * creflav = The client authentication credential flavor (return).
 * cresize = The client authentication credential body length (return).
 * crebody = The client authentication credential body (return).
 * verflav = The client authentication verifier flavor (return).
 * versize = The client authentication verifier body length (return).
 * verbody = The client authentication verifier body (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_rpcx_auth_create (rmix_rpcx_stream_authflav_t  *creflav,
                           unsigned int                 *cresize,
                           char                        **crebody,
                           rmix_rpcx_stream_authflav_t  *verflav,
                           unsigned int                 *versize,
                           char                        **verbody);

/*
 * Authenticates a client at a server.
 *
 * creflav  = The client authentication credential flavor.
 * cresize  = The client authentication credential body length (also return).
 * crebody  = The client authentication credential body (also return).
 * verflav  = The authentication verifier flavor (also return).
 * versize  = The authentication verifier body length (also return).
 * verbody  = The authentication verifier body (also return).
 * authstat = The authentication status (return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_rpcx_auth_authenticate (rmix_rpcx_stream_authflav_t   creflav,
                                 unsigned int                 *cresize,
                                 char                        **crebody,
                                 rmix_rpcx_stream_authflav_t  *verflav,
                                 unsigned int                 *versize,
                                 char                        **verbody,
                                 rmix_rpcx_stream_authstat_t  *authstat);

/*
 * Verifies a server at a client.
 *
 * creflav  = The client authentication credential flavor.
 * cresize  = The client authentication credential body length (also return).
 * crebody  = The client authentication credential body (also return).
 * verflav  = The client authentication verifier flavor.
 * versize  = The client authentication verifier body length (also return).
 * verbody  = The client authentication verifier body (also return).
 * verflav2 = The server authentication verifier flavor.
 * versize2 = The server authentication verifier body length (also return).
 * verbody2 = The server authentication verifier body (also return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_rpcx_auth_verify (rmix_rpcx_stream_authflav_t   creflav,
                           unsigned int                 *cresize,
                           char                        **crebody,
                           rmix_rpcx_stream_authflav_t   verflav,
                           unsigned int                 *versize,
                           char                        **verbody,
                           rmix_rpcx_stream_authflav_t   verflav2,
                           unsigned int                 *versize2,
                           char                        **verbody2);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_RPCX_AUTH_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
